{% extends "create.html" %}

{% block title %}Edit entry{% endblock %}

{% block content_title %}Edit entry{% endblock %}

{% block form_action %}{{ url_for('edit', slug=entry.slug) }}{% endblock %}

{% block save_button %}Save changes{% endblock %}
